/*
 * Jean-Luc Béchennec pour LOCODUINO
 * 
 *  Gestion d'une gare cachée, les objets, la recherche d'un chemin
 */

/*
 * Identifiant des éléments de voie
 */
enum {
  VOIE_ENTREE,
  VOIE_SORTIE,
  AIG_ENTREE_0,
  AIG_ENTREE_1,
  AIG_ENTREE_2,
  AIG_SORTIE_0,
  AIG_SORTIE_1,
  AIG_SORTIE_2,
  VOIE_GARE_0,
  VOIE_GARE_1,
  VOIE_GARE_2,
  VOIE_GARE_3,
  AUCUNE_VOIE /* identifiant spécial */
};

/*
 * Graine aléatoire de qualité
 */
byte getSeed()
{
  byte seed = 0;
  byte i = 8;
  while (i > 0) {
    byte firstBit = analogRead(A0) & 0x01;
    byte secondBit = analogRead(A0) & 0x01;
    if (firstBit != secondBit) {
      seed = (seed << 1) | firstBit;
      i--;
    }
  }
  return seed;
}

/*
 * Affichage en clair du non d'un élément de voie, uniquement pour les aiguillages
 */
void afficheVoie(byte identifiant)
{
  switch (identifiant) {
    case AIG_ENTREE_0:
      Serial.print("Aiguillage E0"); break;
    case AIG_ENTREE_1:
      Serial.print("Aiguillage E1"); break;
    case AIG_ENTREE_2:
      Serial.print("Aiguillage E2"); break;
    case AIG_SORTIE_0:
      Serial.print("Aiguillage S0"); break;
    case AIG_SORTIE_1:
      Serial.print("Aiguillage S1"); break;
    case AIG_SORTIE_2:
      Serial.print("Aiguillage S2"); break;
    case VOIE_GARE_0:
       Serial.print("Voie gare 0"); break;
    case VOIE_GARE_1:
       Serial.print("Voie gare 1"); break;
    case VOIE_GARE_2:
       Serial.print("Voie gare 2"); break;
    case VOIE_GARE_3:
       Serial.print("Voie gare 3"); break;
    default:
      Serial.print("** INCONNU **"); break;
  }
}

/*
 * Classe de base des éléments de voie
 * 
 * Fonctions virtuelles de recherche de chemin et identifiant
 */
class Voie
{
  private:
    byte mIdentifiant;
 
  public:
    Voie(byte identifiant) { mIdentifiant = identifiant; }
    byte idVoie() { return mIdentifiant; }
    virtual bool creeCheminVers(byte identifiant) = 0;
    virtual byte creeCheminVersLibre() = 0;
};

/*
 * Classe pour les voies situées aux extremités
 */
class VoieExtreme : public Voie
{
  private:
    Voie *mVoie;

  public:
    VoieExtreme(byte identifiant) : Voie(identifiant) { mVoie = NULL; }
    void connecte(Voie *voie) { mVoie = voie; }
    virtual bool creeCheminVers(byte identifiant);
    virtual byte creeCheminVersLibre();
};

bool VoieExtreme::creeCheminVers(byte identifiant)
{
  if (identifiant == idVoie()) {
    return true;
  }
  else {
    if (mVoie != NULL) {
      return mVoie->creeCheminVers(identifiant);
    }
    else {
      return false;
    }
  }
}

byte VoieExtreme::creeCheminVersLibre()
{
  if (mVoie != NULL) {
    return mVoie->creeCheminVersLibre();
  }
}


/*
 * Classe pour les voies de gare
 */
class VoieGare : public Voie
{
  private:
    bool mLibre;
    
  public:
    VoieGare(byte identifiant) : Voie(identifiant) { mLibre = true; }
    virtual bool creeCheminVers(byte identifiant);
    virtual byte creeCheminVersLibre();
    bool occupee() { return ! mLibre; }
    void libere()  { mLibre = true; }
};

bool VoieGare::creeCheminVers(byte identifiant)
{
  if (identifiant == idVoie()) {
    return true;
  }
  else {
    return false;
  }
}

byte VoieGare::creeCheminVersLibre()
{
  if (mLibre) {
    Serial.print(F("Occupation de : "));
    afficheVoie(idVoie());
    Serial.println();
    mLibre = false;
    return idVoie();
  }
  else {
    return AUCUNE_VOIE;
  }
}

/*
 * Classe pour les aiguillages
 */
class Aiguillage : public Voie
{
  private:
    Voie *mVoieDroite; /* Voie connectée coté dévié */
    Voie *mVoieDeviee; /* Voie connectée coté droit */
  
  public:
    Aiguillage(byte identifiant) : Voie(identifiant) { mVoieDroite = mVoieDeviee = NULL; }
    void connecteDroite(Voie *voie) { mVoieDroite = voie; }
    void connecteDeviee(Voie *voie) { mVoieDeviee = voie; }
    virtual bool creeCheminVers(byte identifiant);
    virtual byte creeCheminVersLibre();
};

bool Aiguillage::creeCheminVers(byte identifiant)
{
  if (identifiant == idVoie()) {
    return true;
  }
  else {
    if (mVoieDroite != NULL && mVoieDeviee != NULL) {
      /* Uniquement si correctement connecte     */
      /* Cherche la destination cote voie droite */
      if (mVoieDroite->creeCheminVers(identifiant)) {
        /* Destination trouvee cote droit */
        afficheVoie(idVoie());
        Serial.println(": Droit");
        return true;
      }
      else {
        /* Cherche la destination cote voie deviee */
        if (mVoieDeviee->creeCheminVers(identifiant)) {
          /* destination trouvee cote devie */
          afficheVoie(idVoie());
          Serial.println(": Devie");
          return true;
        }
        else {
          /* destination non trouvee */
          return false;
        }
      }
    }
  }
}

byte Aiguillage::creeCheminVersLibre()
{
  byte voie;
  if (mVoieDroite != NULL && mVoieDeviee != NULL) {
    /* Uniquement si correctement connecte     */
    /* Cherche la destination cote voie droite */
    voie = mVoieDroite->creeCheminVersLibre();
    if (voie != AUCUNE_VOIE) {
      /* Destination trouvee cote droit */
      afficheVoie(idVoie());
      Serial.println(F(": Droit"));
      return voie;
    }
    else {
      /* Cherche la destination cote voie deviee */
      voie = mVoieDeviee->creeCheminVersLibre();
      if (voie != AUCUNE_VOIE) {
        /* destination trouvee cote devie */
        afficheVoie(idVoie());
        Serial.println(F(": Devie"));
        return voie;
      }
      else {
        /* destination non trouvee */
        return AUCUNE_VOIE;
      }
    }
  }
}

/*
 * Instantiation des objets de la gare
 */
VoieExtreme voieEntree(VOIE_ENTREE);
VoieExtreme voieSortie(VOIE_SORTIE);
Aiguillage aiguillageE0(AIG_ENTREE_0);
Aiguillage aiguillageE1(AIG_ENTREE_1);
Aiguillage aiguillageE2(AIG_ENTREE_2);
Aiguillage aiguillageS0(AIG_SORTIE_0);
Aiguillage aiguillageS1(AIG_SORTIE_1);
Aiguillage aiguillageS2(AIG_SORTIE_2);

const int TAILLE_GARE = 4;

VoieGare voieGare[TAILLE_GARE] = {
   VoieGare(VOIE_GARE_0),
   VoieGare(VOIE_GARE_1),
   VoieGare(VOIE_GARE_2),
   VoieGare(VOIE_GARE_3)
};


void testLibre()
{
  if (voieEntree.creeCheminVersLibre() != AUCUNE_VOIE)
    Serial.println(F("TROUVE"));
  else
    Serial.println(F("NON TROUVE"));
}

/*
 * Sélectionne aléatoirement un des trains des voies occupées
 */
byte selectionneTrain()
{
  byte voiesOccupees[TAILLE_GARE];
  byte indexVoieOccupee = 0;
  /* Parcoure les voies de la gare */
  for (byte indexVoieGare = 0; indexVoieGare < TAILLE_GARE; indexVoieGare++) {
    if (voieGare[indexVoieGare].occupee()) {
      /* Une voie occupee est trouvée, on ajoute son identifiant à notre tableau */
      voiesOccupees[indexVoieOccupee++] = voieGare[indexVoieGare].idVoie();
    }
  }
  if (indexVoieOccupee > 0) {
    /* Au moins une voie occupée a été trouvée */
    /* Tire au hasard une voie occupée */
    byte voie = random(0, indexVoieOccupee);
    /* et retourne la voie */
    return voiesOccupees[voie];   
  }
  else {
    /* la gare est vide ! */
    return AUCUNE_VOIE;
  }
}

/*
 * Libère la voie dont l'identifiant est passé en argument
 */
void libereVoieGare(byte identifiant)
{
  for (byte indexVoieGare = 0; indexVoieGare < TAILLE_GARE; indexVoieGare++) {
    if (voieGare[indexVoieGare].idVoie() == identifiant) {
      /* La voie d'indentifiant correspondant est trouvée */
      voieGare[indexVoieGare].libere();
      /* Comme on a trouvé, on sort de la boucle */
      break;
    }
  } 
}

/*
 * Initialisation et test
 */
void setup() {
  /* Connexion des elements de voie du grill d'entree */
  voieEntree.connecte(&aiguillageE0);
  aiguillageE0.connecteDroite(&voieGare[0]);
  aiguillageE0.connecteDeviee(&aiguillageE1);
  aiguillageE1.connecteDroite(&aiguillageE2);
  aiguillageE1.connecteDeviee(&voieGare[1]);
  aiguillageE2.connecteDroite(&voieGare[3]);
  aiguillageE2.connecteDeviee(&voieGare[2]);
  /* Connexion des elements de voie du grill de sortie */
  voieSortie.connecte(&aiguillageS0);
  aiguillageS0.connecteDroite(&voieGare[3]);
  aiguillageS0.connecteDeviee(&aiguillageS1);
  aiguillageS1.connecteDroite(&aiguillageS2);
  aiguillageS1.connecteDeviee(&voieGare[2]);
  aiguillageS2.connecteDroite(&voieGare[0]);
  aiguillageS2.connecteDeviee(&voieGare[1]);

  Serial.begin(9600);
  Serial.println("***** TEST de chemin vers identifiant");
  /* Test des chemins */
  Serial.println("===== Test de l'entree");
  Serial.println("Voie gare 0");
  voieEntree.creeCheminVers(VOIE_GARE_0);
  Serial.println("Voie gare 1");
  voieEntree.creeCheminVers(VOIE_GARE_1);
  Serial.println("Voie gare 2");
  voieEntree.creeCheminVers(VOIE_GARE_2);
  Serial.println("Voie gare 3");
  voieEntree.creeCheminVers(VOIE_GARE_3);
  
  Serial.println("===== Test de la sortie");
  Serial.println("Voie gare 0");
  voieSortie.creeCheminVers(VOIE_GARE_0);
  Serial.println("Voie gare 1");
  voieSortie.creeCheminVers(VOIE_GARE_1);
  Serial.println("Voie gare 2");
  voieSortie.creeCheminVers(VOIE_GARE_2);
  Serial.println("Voie gare 3");
  voieSortie.creeCheminVers(VOIE_GARE_3);

  Serial.println("***** TEST de chemin vers libre");
  testLibre();    
  testLibre();    
  testLibre();    
  testLibre();    
  testLibre();

  /* Initialise le générateur de nombres aléatoires */
  randomSeed(getSeed());
  
  Serial.println("***** TEST de vidange de la gare");
  while (true)
  {
    byte identifiant = selectionneTrain();
    if (identifiant != AUCUNE_VOIE) {
      Serial.print("Selectionne le train de : ");
      afficheVoie(identifiant);
      Serial.println();
      voieSortie.creeCheminVers(identifiant);
      libereVoieGare(identifiant);
    }
    else {
      break;    
    }
  }
  Serial.println("TERMINE");
}

void loop() {
  /* Rien pour l'instant */
}
